"""Pandantic is a library for validating and serializing data using Pydantic and Pandas."""
from pandantic.basemodel import PandanticBaseModel as BaseModel
from pandantic.types_pandantic import Optional  # type: ignore
